# 弾幕系シューティングゲーム「弾幕」 by Kenichi Takebe
import pgzrun, math, random

WIDTH, HEIGHT = 1000, 600 # 画面の幅と高さを設定

class Bullet(Actor): # Bullet（弾）クラスを定義。Actorを継承

  def __init__(self, image): # 初期化関数を定義
    self.power = 0
    super().__init__(image, (-500, 0)) # Actorの__init__関数を実行

  def draw(self): # 描画関数を定義
    if self.power == 0: return
    super().draw() # Actorのdraw関数を実行

  def shot(self, pos, vx, vy): # 弾の発射時に実行する関数を定義
    self.power, self.pos, self.vx, self.vy = 1, pos, vx, vy

  def move(self): # 弾の移動関数を定義
    if self.power == 0: return
    self.x, self.y = (self.x + self.vx), (self.y + self.vy)
    self.angle += 10
    # 画面の外に出たか？
    if not (-50 < self.y < HEIGHT+50 and -50 < self.x < WIDTH+50):
      self.power, self.angle, self.pos = 0, 0, (-500, 0)

# メイン処理
ship = Actor("ship", (WIDTH/2, HEIGHT/2 + 100)) # 自機を作成
enemy = Actor("enemy", (-500, 0)) # 敵を作成

count,ecount,score,num_ebullet,ship.life,bgx = 0,0,0,1000,100,0

# 三角関数のルックアップテーブルを作成
cos_lut = [math.cos(i*3.14159/180) for i in range(360*5)]
sin_lut = [math.sin(i*3.14159/180) for i in range(360*5)]

# 弾（自弾と敵弾）を管理するリスト
bullet = [Bullet("sbullet") for i in range(10)] # 自弾を作成
for i in range(8): bullet.append(Bullet("ebullet3")) # 敵弾を作成
for i in range(int(num_ebullet / 3)): # 敵弾を作成
  for j in range(3): bullet.append(Bullet("ebullet" + str(j)))

def draw(): # 描画関数
  screen.blit("background", (bgx, 0)) # 背景を描画
  # スコアとライフを描画
  screen.draw.text("SCORE : " + str(score) + "\nLIFE : " + \
  str(ship.life), topleft=(10, 10), color="WHITE", fontsize=36)
  for b in bullet: b.draw() # 弾を描画
  enemy.draw() # 敵を描画
  ship.draw() # 自機を描画
  if ship.life == 0: # ゲームオーバーか？
    screen.draw.text("GAME OVER", center=(WIDTH/2, HEIGHT/2), \
    color="WHITE",gcolor="BLACK",fontsize=80,owidth=1,ocolor="GRAY")

def trajectory_change():
  a = int(math.atan2(ship.y - bullet[10].y, ship.x - bullet[10].x) \
  * 180 / 3.14159)
  for i in range(10, 18):
    bullet[i].vx, bullet[i].vy = cos_lut[a] * 5, sin_lut[a] * 5

def update(): # 更新関数
  global count, ecount, num_ebullet, score, ship, bgx
  count, ecount, bgx = (count + 1) % 360, (ecount + 0.1) % 360, 0
  # 敵弾の総数を増やす
  if count == 359: num_ebullet += random.randint(100, 200) 
  # num_ebulletが1017以上になったらnum_ebulletを18にする
  if num_ebullet >= 1017: num_ebullet = 18

  for b in bullet: b.move() # 弾を動かす
  # 敵を動かす
  enemy.x = WIDTH/2 + math.cos(ecount * 9.42477 / 180) * 300 
  enemy.y = HEIGHT/2 + math.sin(ecount * 12.56636 / 180) * 100

  for i in range(10, num_ebullet): # 敵弾と自機の当たり判定を行う
    if bullet[i].power and bullet[i].collidepoint(ship.pos):
      ship.life, bgx = (ship.life - 1), 15
      if ship.life < 0: ship.pos, ship.life = (-500, 0), 0

  for i in range(10): # 自弾と敵の当たり判定を行う
    if bullet[i].power and bullet[i].colliderect(enemy):
      score, bullet[i].power = (score + 1), 0 

  if count == 180: # 円状狙い撃ち弾を発射
    for i, a in zip(range(10, 18), (0,45,90,135,180,225,270,315)):
      bullet[i].shot(enemy.pos, cos_lut[a]*2, sin_lut[a]*2)
    clock.schedule(trajectory_change, 0.5)
  if count < 120 and count % 5 == 0: # 12WAY弾を発射
    for a in (0, 30, 60, 90, 120, 150, 180, 210, 240, 270, 300, 330):
      for i in range(18, num_ebullet, 3):
        if bullet[i].power == 0:
          bullet[i].shot(enemy.pos, cos_lut[a]*3, sin_lut[a]*3)
          break
  if count % 4 == 0:  # 3WAYの渦巻き弾を発射
    for a in (0, 120, 240):
      for i in range(19, num_ebullet, 3):
        if bullet[i].power == 0:
          vx = cos_lut[count * 4 + a] * 3
          vy = sin_lut[count * 4 + a] * 3
          bullet[i].shot(enemy.pos, vx, vy)
          break
  if count > 200 and count % 30 == 0: # 円状弾を発射
    a = 5
    for i in range(20, num_ebullet, 3):
      if bullet[i].power == 0:
        bullet[i].shot(enemy.pos, cos_lut[a]*3, sin_lut[a]*3)
        a += random.randint(5, 20) 
        if a > 360: break

  if ship.life == 0: return # ゲームオーバーだったらここで終了

  # 自機の移動処理
  if keyboard.up and ship.y > 0: ship.y -= 3 
  if keyboard.down and ship.y < HEIGHT: ship.y += 3
  if keyboard.left and ship.x > 0: ship.x -= 3
  if keyboard.right and ship.x < WIDTH: ship.x += 3
  if keyboard.space and count % 4 == 0: # 自弾の発射処理
    for i in range(10):
      if bullet[i].power == 0:
        a=int(math.atan2(enemy.y-ship.y, enemy.x-ship.x)*180/3.14159)
        bullet[i].shot(ship.pos, cos_lut[a] * 8, sin_lut[a] * 8)
        break

pgzrun.go()